this.build_blacksmith_event <- this.inherit("scripts/events/event", {
	m = {
		Blacksmith = null,
		Town = null,
		Costs = 5000,
	},
	function create()
	{
		this.m.ID = "event.build_blacksmith";
		this.m.Title = "At %townname%";
		this.m.Cooldown = 1.0 * this.World.getTime().SecondsPerDay;
		this.m.Screens.push({
			ID = "A",
			Text = "[img]gfx/ui/events/event_20.png[/img]You find %blacksmith% staying and looking at this small and quiet town. He says that misses his home and village he from, being mercenary is good and fast way to earn enough crowns and get his own blacksmith. May be in such place as this and villagers will be glad to have his service. Just few years and required sum of 5.000 crowns will be saved - for peaceful life without blood and deaths. You look at him, if only he will survive till this time... But, damn, you know that you can make his dream come true!",
			Image = "",
			List = [],
			Characters = [],
			Options = [
				{
					Text = "I will give you all 5.000 crowns just now.",
					function getResult( _event )
					{
						return "B";
					}
				},
				{
					Text = "Company still need you, stay with us.",
					function getResult( _event )
					{
						return 0;
					}
				}
			],
			function start( _event )
			{
				this.Characters.push(_event.m.Blacksmith.getImagePath());
			}
		});
		this.m.Screens.push({
			ID = "B",
			Text = "[img]gfx/ui/events/event_20.png[/img]Few minutes late all company bids farewell with %blacksmith%. That is really encouraging moment, knowing that at least one of us has everything what he wants. New life, family, home and job which helps people without smashing heads. %blacksmith% feels very happy and you also understand that didn\'t waste money.",
			Image = "",
			List = [],
			Characters = [],
			Options = [
				{
					Text = "Don\'t forget us, we are your family forever.",
					function getResult( _event )
					{
						return 0;
					}
				}
			],
			function start( _event )
			{
				this.Characters.push(_event.m.Blacksmith.getImagePath());
				local isWeapon = this.Math.rand(1, 100) > 50;

				this.World.Assets.addMoney(-_event.m.Costs);
				if (isWeapon)
				{
					_event.m.Town.addBuilding(this.new("scripts/entity/world/settlements/buildings/armorsmith_building"));
				}
				else
				{
					_event.m.Town.addBuilding(this.new("scripts/entity/world/settlements/buildings/weaponsmith_building"));
				}

				_event.m.Blacksmith.getItems().transferToStash(this.World.Assets.getStash());
				this.World.getPlayerRoster().remove(_event.m.Blacksmith);
				_event.m.Blacksmith = null;

				this.List.push({
					id = 10,
					icon = "ui/icons/asset_money.png",
					text = "You spend [color=" + this.Const.UI.Color.NegativeEventValue + "]" + _event.m.Costs + "[/color] Crowns"
				});

				this.List.push({
					id = 10,
					icon = "ui/icons/special.png",
					text = _event.m.Blacksmith.getName() + " left company"
				});

				if (isWeapon)
				{
					this.List.push({
						id = 10,
						icon = "ui/icons/buildings/weaponsmith.png",
						text = _event.m.Town.getName() + " now has weaponsmith"
					});

					this.World.FactionManager.getFaction(_event.m.Town.getFactions()[0]).addPlayerRelation(this.Const.World.Assets.ReputationOnAmbition, "You built weaponsmith");
				}
				else
				{
					this.List.push({
						id = 10,
						icon = "ui/icons/buildings/armorsmith.png",
						text = _event.m.Town.getName() + " now has armorsmith"
					});

					this.World.FactionManager.getFaction(_event.m.Town.getFactions()[0]).addPlayerRelation(this.Const.World.Assets.ReputationOnAmbition, "You built armorsmith");
				}
			}
		});
	}

	function onUpdateScore()
	{
		if (this.World.Assets.getMoney() < this.m.Costs)
		{
			return;
		}

		local towns = this.World.EntityManager.getSettlements();
		local nearTown = false;
		local town;
		local playerTile = this.World.State.getPlayer().getTile();

		foreach( t in towns )
		{
			if (t.getTile().getDistanceTo(playerTile) > 3 || !t.isAlliedWithPlayer())
			{
				continue;
			}

			if (t.getSize() > 1 || t.isMilitary())
			{
				continue;
			}

			if (t.hasBuilding("building.temple") || t.hasBuilding("building.armorsmith") || t.hasBuilding("building.weaponsmith") || t.hasBuilding("building.fletcher"))
			{
				continue;
			}

			nearTown = true;
			town = t;
			break;
		}

		if (!nearTown)
		{
			return;
		}

		local brothers = this.World.getPlayerRoster().getAll();
		local candidates = [];

		foreach( bro in brothers )
		{
			if (bro.getBackground().getID() == "background.apprentice")
			{
				candidates.push(bro);
			}
		}

		if (candidates.len() == 0)
		{
			return;
		}

		this.m.Blacksmith = candidates[this.Math.rand(0, candidates.len() - 1)];
		this.m.Town = town;
		this.m.Score = 500;
	}

	function onPrepare()
	{
	}

	function onPrepareVariables( _vars )
	{
		_vars.push([
			"blacksmith",
			this.m.Blacksmith.getNameOnly()
		]);
		_vars.push([
			"townname",
			this.m.Town.getName()
		]);
	}

	function onClear()
	{
		this.m.Blacksmith = null;
		this.m.Town = null;
	}
});
